from bevy import inject, dependency

from wordlette.cms.theming import Template, ThemeManager
from wordlette.middlewares.router_middleware import RouteManager
from wordlette.requests import Request
from wordlette.routes import Route
from wordlette.state_machines import State


class _SetupRoute(Route):
    class __metadata__:
        abstract = True
        registry = set()


class Index(_SetupRoute):
    path = "/"

    async def get(self, request: Request.Get):
        return Template(
            "index.html", title="Wordlette", subtitle="Setting Up Your Site"
        )


class Setup(State):
    @inject
    async def enter_state(
        self,
        route_manager: RouteManager = dependency(),
        theme_manager: ThemeManager = dependency(),
    ):
        theme_manager.set_theme(theme_manager.wordlette_res / "themes" / "setup")
        _SetupRoute.register_routes(route_manager.router)



from typing import BinaryIO, Any

try:
    import tomlkit as toml

    can_write = True
except ImportError:
    can_write = False

    try:
        import tomllib as toml
    except ImportError:
        toml = None

from wordlette.configs.handlers import Serializable, ConfigHandler


class TomlHandler(ConfigHandler):
    extensions = {"toml"}

    def load(self, file: BinaryIO) -> dict[str, Any]:
        if not toml:
            raise ImportError("tomlkit or tomllib is required to load TOML files.")

        return toml.load(file)

    def write(self, data: dict | Serializable, file: BinaryIO):
        if not can_write:
            raise ImportError("tomlkit is required to write TOML files.")

        def write(d):
            file.write(toml.dumps(d).encode())

        match data:
            case dict():
                write(data)

            case Serializable():
                write(data.to_dict())

            case _:
                raise TypeError(
                    f"Cannot write type {type(data)!r} into a TOML file. Must be a dict or adhere to the "
                    f"`{Serializable.__module__}.{Serializable.__qualname__}` protocol."
                )

    @classmethod
    def supported(cls) -> bool:
        return toml is not None



from functools import partial
from types import MethodType
from typing import Callable, Awaitable

from starlette.types import Scope, Send

from wordlette.requests import Request


class Middleware:
    def __init__(self, call_next: Callable[[Scope, Request, Send], Awaitable[None]]):
        super().__init__()
        self.call_next = call_next

    async def next(self):
        """Calls the next middleware in the stack. A context aware method is injected at runtime that can correctly call
        the next middleware."""

    async def run(self, scope: Scope, request: Request, send: Send):
        await self.next()

    async def __call__(self, scope: Scope, request: Request, send: Send):
        await call_with_extended_self(
            self.run,
            (scope, request, send),
            {},
            next=partial(self.call_next, scope, request, send),
        )


def call_with_extended_self(method: MethodType, args, kwargs, **extensions):
    self = method.__self__
    extended_type = type("ExtendedType", (self.__class__,), extensions)
    extended_self = object.__new__(extended_type)
    extended_self.__dict__ = self.__dict__
    return method.__func__(extended_self, *args, **kwargs)



import logging
from abc import ABCMeta, abstractmethod
from typing import Generic, Type, TypeVar, ParamSpec, Callable, Awaitable

from wordlette.state_machines.predicates import always

logger = logging.getLogger("StateMachine")

_T = TypeVar("_T")
_P = ParamSpec("_P")


class RequestCycle:
    pass


class Transition:
    __match_args__ = ("state", "to_state", "predicate")

    def __init__(
        self,
        state: "Type[State[_T]]",
        to_state: "Type[State[_T]]",
        predicate: Callable[[], Awaitable[bool]],
    ):
        self.state = state
        self.to_state = to_state
        self.predicate = predicate


class StateABCMeta(ABCMeta):
    def __repr__(cls):
        return f"<class State:{cls.__name__}>"


class State(Generic[_T], metaclass=StateABCMeta):
    transitions = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.transitions = cls.transitions.copy()

    def cycle(self) -> RequestCycle:
        return RequestCycle()

    @abstractmethod
    async def enter_state(self) -> RequestCycle | None:
        ...

    async def exit_state(self):
        return

    def __repr__(self):
        return f"<State:{self.__class__.__name__}>"

    @classmethod
    def goes_to(
        cls, state: "Type[State[_T]]", when: Callable[_P, Awaitable[bool]] = always
    ) -> Transition:
        return Transition(cls, state, when)


class NullState(State[_T]):
    async def enter_state(self) -> None:
        return None


class InitialState(NullState[_T]):
    pass



import inspect
from collections.abc import Set
from types import MethodType, UnionType
from typing import (
    Any,
    Callable,
    Generic,
    Iterator,
    ParamSpec,
    Type,
    TypeVar,
    cast,
    get_args,
)

from starlette.types import Scope, Receive, Send

from wordlette.core.exceptions import MissingRoutePath, NoRouteHandlersFound
from wordlette.requests import Request
from wordlette.routers import Router
from wordlette.routes.exception_contexts import ExceptionHandlerContext
from wordlette.utils.apply import apply
from wordlette.utils.match_types import TypeMatchable
from wordlette.utils.options import Option

P = ParamSpec("P")
RequestType = TypeVar("RequestType", bound=Request)
ExceptionType = TypeVar("ExceptionType", bound=Exception)


class MethodsCollection(Set[Type[Request]]):
    def __init__(self, *methods: Type[Request]):
        self._methods = set(methods)

    @property
    def names(self):
        return [method.name for method in self]

    def __contains__(self, request: Request | Type[Request]) -> bool:
        request_type = type(request) if isinstance(request, Request) else request
        return request_type in self._methods

    def __iter__(self) -> Iterator[Type[Request]]:
        return iter(self._methods)

    def __len__(self) -> int:
        return len(self._methods)

    def __repr__(self):
        return f"<{type(self).__name__} {{{', '.join(sorted(method.name for method in self))}}}>"


class RouteMetadata:
    abstract: bool
    registry: "set[Type[Route]]"
    error_handlers: dict[Type[Exception], Callable[[Exception], Any]]
    request_handlers: dict[Type[Request], Callable[[Request], Any]]


class _RouteMetadata:
    class __metadata__(RouteMetadata):
        abstract = True
        registry = set()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if _RouteMetadata in cls.__bases__:
            return

        cls._setup_metadata()

    @classmethod
    def _create_new_metadata_object(cls):
        super_route = get_super_route(cls)
        cls.__metadata__ = create_metadata_type(
            super_route.__metadata__,
            abstract=False,
            request_handlers={},
            error_handlers={},
        )

    @classmethod
    def _setup_metadata(cls):
        if cls.__metadata__ is get_super_route(cls).__metadata__:
            cls._create_new_metadata_object()

        elif not issubclass(cls.__metadata__, RouteMetadata):
            cls._transform_metadata_to_metadata_type()

    @classmethod
    def _transform_metadata_to_metadata_type(cls):
        super_route = get_super_route(cls)

        if not hasattr(cls.__metadata__, "request_handlers"):
            cls.__metadata__.request_handlers = {}

        if not hasattr(cls.__metadata__, "error_handlers"):
            cls.__metadata__.error_handlers = {}

        cls.__metadata__ = create_metadata_type(
            cls.__metadata__, super_route.__metadata__
        )


class RouteMCS(type):
    methods: MethodsCollection
    name: str
    path: str

    def __repr__(cls):
        path, name, methods = cls.path, cls.name, cls.methods.names
        return f"<Route:{cls.__name__} {path=} {name=} {methods=}>"


class Route(Generic[RequestType], _RouteMetadata, TypeMatchable, metaclass=RouteMCS):
    """The route type handles all the magic that allows routes to be defined without any decorator boilerplate. It
    provides the instrumentation necessary for simple request handling by method type using type annotations. It also
    provides a simple exception handling mechanism using the same approach.

    To declare a request handler create an async method with whatever name you think is appropriate. Then all you have
    to do is set the type annotion of the first parameter to the request type your handler should handle. For example:
    ```python
    class MyRoute(Route):
        async def index(self, request: Request.Get):
            return PlainTextResponse("Hello World!")
    ```
    Handling errors is just as simple and follows the exact same process. The exception type will match all exceptions
    that pass a subclass check, so it should function the same as an equivalent try/except. For example:
    ```python
    class MyRoute(Route):
        ...
        async def handle_exception(self, error: Exception):
            return PlainTextResponse("An error occurred!")
    ```
    """

    methods: MethodsCollection
    name: str
    path: str

    def __init_subclass__(cls, **kwargs):
        """Setup the route class building the route meta, scanning for request & error handlers, and ensuring that
        concrete routes have a path and route handlers."""
        super().__init_subclass__(**kwargs)
        cls._setup_route()
        cls._scan_for_handlers()
        cls._validate_route_object()
        cls._register_route()

    async def __call__(self, scope, receive, send):
        """Handle a request, calling the appropriate handler function and capturing any handleable exceptions."""
        async with ExceptionHandlerContext(self) as error_handler:
            await self._handle(scope, receive, send)

        if error_handler:
            await error_handler(scope, receive, send)

    async def _handle(self, scope: Scope, receive: Receive, send: Send):
        request = Request.factory(scope)
        if type(request) not in self.__metadata__.request_handlers:
            raise Exception(
                f"Request type {type(request)} not handled by {self.__class__.__qualname__}"
            )

        handler_registry = self.__metadata__.request_handlers
        request_type = type(request)
        response = await handler_registry[request_type](self, request)
        await response(scope, receive, send)

    @classmethod
    def register_routes(cls, router: Router):
        """Register all routes in the route registry with the given router."""
        apply(router.add_route).to(cls.__metadata__.registry)

    @classmethod
    def _add_handlers(cls, function, handler_types, container):
        for handler_type in handler_types:
            container[handler_type] = function

    @classmethod
    def _register_handlers(cls, function, *handler_types):
        if all(issubclass(handler, Request) for handler in handler_types):
            cls._add_handlers(
                function, handler_types, cls.__metadata__.request_handlers
            )

        elif all(issubclass(handler, Exception) for handler in handler_types):
            cls._add_handlers(function, handler_types, cls.__metadata__.error_handlers)

    @classmethod
    def _register_route(cls):
        if cls.__metadata__.abstract:
            return

        cls.__metadata__.registry.add(cls)

    @classmethod
    def _scan_for_handlers(cls):
        for function in get_functions(cls):
            handler_type = get_handler_type(function)
            match handler_type:
                case Option.Value(type() as ht) if issubclass(ht, (Request, Exception)):
                    cls._register_handlers(function, ht)

                case Option.Value(UnionType() as ht):
                    cls._register_handlers(function, *get_args(ht))

        cls.methods = MethodsCollection(*cls.__metadata__.request_handlers)

    @classmethod
    def _setup_route(cls):
        if cls.__metadata__.abstract:
            return

        if not hasattr(cls, "name"):
            cls.name = cls.__qualname__.casefold()

    @classmethod
    def _validate_route_object(cls):
        if cls.__metadata__.abstract:
            return

        if not hasattr(cls, "path"):
            raise MissingRoutePath(
                f"Route subclass {cls.__qualname__} is missing a path attribute."
            )

        if not cls.methods:
            raise NoRouteHandlersFound(
                f"Route subclass {cls.__qualname__} has no request handlers."
            )


def create_metadata_type(
    *bases: Type[RouteMetadata] | Type, **properties: Any
) -> Type[RouteMetadata]:
    return cast(
        Type[RouteMetadata],
        type(
            "__metadata__",
            bases,
            properties,
        ),
    )


def get_functions(cls):
    for name, attr in vars(cls).items():
        if (
            callable(attr)
            and not isinstance(attr, MethodType)
            and not name.startswith("_")
        ):
            yield attr


def get_handler_type(function):
    arg_spec = inspect.getfullargspec(unwrap(function))
    arg_type = (
        arg_spec.annotations.get(arg_spec.args[1]) if len(arg_spec.args) > 1 else None
    )
    return Option.Value(arg_type) if arg_type else Option.Null()


def get_super_route(cls):
    for base in cls.__bases__:
        if base is Route or issubclass(base, Route):
            return base


def unwrap(function):
    while hasattr(function, "__wrapped__"):
        function = function.__wrapped__

    return function



class EventMCS(type):
    def __repr__(self):
        return f"<class Event:{self.__name__}>"


class Event(metaclass=EventMCS):
    __event_count__: int
    __event_name__: str

    def __init_subclass__(cls, **kwargs):
        cls.__event_count__ = 0
        if not hasattr(cls, "__event_name__"):
            cls.__event_name__ = cls.__name__

    def __new__(cls, *args, **kwargs):
        event_id, cls.__event_count__ = cls.__event_count__, cls.__event_count__ + 1
        instance = super().__new__(cls)
        instance.__event_count__ = event_id
        return instance

    def __repr__(self):
        return f"<Event:{self.__event_name__} {self.__event_count__}>"



from typing import Any, Protocol, runtime_checkable, BinaryIO


@runtime_checkable
class Serializable(Protocol):
    def to_dict(self) -> dict[str, Any]:
        ...


@runtime_checkable
class ConfigHandler(Protocol):
    extensions: set[str]

    def load(self, file: BinaryIO) -> dict[str, Any]:
        ...

    def write(self, data: dict | Serializable, file: BinaryIO):
        ...



import logging
from functools import reduce
from typing import (
    TypeAlias,
    Callable,
    Awaitable,
    Sequence,
    Type,
    cast,
    overload,
    runtime_checkable,
    Protocol,
)

from bevy import get_repository
from starlette.responses import PlainTextResponse
from starlette.types import Receive, Send, Scope, Message, ASGIApp

from wordlette.core.events import (
    LifespanStartupEvent,
    LifespanShutdownEvent,
)
from wordlette.core.exceptions import InvalidExtensionOrConstructor
from wordlette.events import Observer, Observable
from wordlette.extensions import Extension
from wordlette.middlewares import Middleware
from wordlette.state_machines import StateMachine
from wordlette.utils.apply import apply

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("App")

_MiddlewareConstructor: TypeAlias = Callable[[ASGIApp], Middleware] | Type[Middleware]


@runtime_checkable
class CallableProtocol(Protocol):
    def __call__(self, *args, **kwargs):
        ...


class Sender:
    def __init__(self, send: Send):
        self.sent = False
        self.send = send

    def __call__(self, message: Message) -> Awaitable[None]:
        self.sent = True
        return self.send(message)


class WordletteApp(Observable):
    def __init__(
        self,
        *,
        extensions: Sequence[Callable[[], Extension]] = (),
        middleware: Sequence[_MiddlewareConstructor],
        state_machine: StateMachine,
    ):
        self._update_repository()
        self._extensions: dict[str, Extension] = {}
        self._middleware_stack: ASGIApp = self._build_middleware_stack(middleware)
        self._state_machine = state_machine

        self._state_machine.__event_dispatch__.propagate_to(
            self.__event_dispatch__.emit
        )
        self._build_extensions(extensions)

    @overload
    def add_extension(self, name: str, extension: Extension):
        ...

    @overload
    def add_extension(self, name: str, extension_constructor: Callable[[], Extension]):
        ...

    @overload
    def add_extension(self, extension_constructor: Callable[[], Extension]):
        ...

    @overload
    def add_extension(self, extension: Extension):
        ...

    def add_extension(self, *args):
        match args:
            case (str() as name, Extension() as extension):
                pass

            case (str() as name, CallableProtocol() as extension_constructor):
                extension = extension_constructor()

            case (Extension(name) as extension,):
                pass

            case (CallableProtocol() as extension_constructor,):
                extension = extension_constructor()
                name = extension.name

            case _:
                raise InvalidExtensionOrConstructor(
                    "You must pass an extension or an extension constructor, got {args}. A valid constructor can be "
                    "any type or callable that returns an extension object."
                )

        self._add_extension(name, extension)

    async def handle_lifespan(self, scope: Scope, receive: Receive, send: Send):
        running = True
        while running:
            match await receive():
                case {"type": "lifespan.startup"}:
                    logger.debug("Lifespan startup event")
                    await self.emit(LifespanStartupEvent(scope))
                    await send({"type": "lifespan.startup.complete"})

                case {"type": "lifespan.shutdown"}:
                    logger.debug("Lifespan shutdown event")
                    await self.emit(LifespanShutdownEvent(scope))
                    await send({"type": "lifespan.shutdown.complete"})
                    running = False

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if not self._state_machine.started:
            try:
                await self._state_machine.cycle()
            except Exception as e:
                logger.exception("Error while starting state machine", exc_info=True)
                raise e

        match scope:
            case {"type": "lifespan"}:
                await self.handle_lifespan(scope, receive, send)

            case _:
                try:
                    await self._middleware_stack(scope, receive, Sender(send))
                except Exception as e:
                    logger.exception("Error in middleware stack", exc_info=True)
                    raise e

    async def _500_response(self, scope: Scope, receive: Receive, send: Sender):
        if send.sent:
            return

        await PlainTextResponse("Internal Server Error", status_code=500)(
            scope, receive, send
        )

    def _build_extensions(self, extension_constructors):
        apply(self.add_extension).to(extension_constructors)

    def _build_middleware_stack(self, middleware_constructors) -> ASGIApp:
        def middleware_factory(previous, current) -> ASGIApp:
            return current(previous)

        return reduce(
            middleware_factory,
            middleware_constructors,
            cast(ASGIApp, self._500_response),
        )

    def _update_repository(self):
        repo = get_repository()
        repo.set(WordletteApp, self)

    def _add_extension(self, name: str, extension: Extension):
        self._extensions[name] = extension
        get_repository().set(type(extension), extension)
        if isinstance(extension, Observer):
            extension.observe(self)



from wordlette.configs.json_handlers import JsonHandler
from wordlette.configs.models import ConfigModel
from wordlette.configs.toml_handlers import TomlHandler
from wordlette.configs.yaml_handlers import YamlHandler



from types import MethodType
from typing import Type, get_origin, get_args, Union, Any

from wordlette.events import Observable
from wordlette.events.dispatch import Callback
from wordlette.events.dispatchable import Dispatchable
from wordlette.events.events import Event


def _unwrap(func):
    while hasattr(func, "__wrapped__"):
        func = func.__wrapped__

    return func


def _get_first_arg_types(func) -> tuple[Type[Any], ...]:
    if not hasattr(func, "__annotations__"):
        return ()

    if isinstance(func, MethodType):
        return ()

    annotations = list(func.__annotations__.values())
    if len(annotations) < 1:
        return ()

    if get_origin(annotations[0]) is Union:
        return get_args(annotations[0])

    return (annotations[0],)


class Observer(Dispatchable):
    __event_listeners__: dict[Type[Event], Callback]

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        for event_type, listener in cls.__event_listeners__.items():
            instance.__event_dispatch__.listen(
                event_type, MethodType(listener, instance)
            )

        return instance

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.__event_listeners__ = {}
        for name in dir(cls):
            if not name.startswith("_"):
                func = getattr(cls, name)
                arg_types = _get_first_arg_types(_unwrap(func))
                for arg_type in arg_types:
                    if isinstance(arg_type, type) and issubclass(arg_type, Event):
                        cls.__event_listeners__[arg_type] = func

    def observe(self, observable: Observable | Type[Observable]):
        observable.__event_dispatch__.propagate_to(self.__event_dispatch__.emit)



from typing import Type, TypeVar, cast

T = TypeVar("T")


class IsSubclassMCS(type):
    match_type: Type

    def __instancecheck__(self, instance):
        return isinstance(instance, type) and issubclass(instance, self.match_type)


def match_type(cls) -> IsSubclassMCS:
    return cast(
        Type[T], IsSubclassMCS(f"IsSubclass_{cls.__name__}", (), {"match_type": cls})
    )


class TypeMatchable:
    matches_type: IsSubclassMCS

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.matches_type = match_type(cls)



from typing import Any, Iterable


class Apply:
    def __init__(self, *callables):
        self.callables = callables

    def to(self, *iterables: Iterable[Any]):
        for args in zip(*iterables):
            for callable_ in self.callables:
                callable_(*args)


def apply(*callables) -> Apply:
    return Apply(*callables)



from typing import Type


class ExceptionHandlerContext:
    """When an exception occurs in a route, the exception handler context finds a matching handler on the route. If no
    handler is found the exception will be allowed to propagate. Calling the context object after it has found a handler
    runs the handler and its response, sending the response to the client."""

    def __init__(self, route):
        self.route = route
        self.exception = None
        self.handler = None

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if handler := self._find_error_handler(exc_type):
            self.handler = handler
            self.exception = exc_type(exc_val)
            self.exception.with_traceback(exc_tb)
            return True

    def __bool__(self):
        return self.exception is not None

    async def __call__(self, scope, receive, send):
        response = await self.handler(self.route, self.exception)
        await response(scope, receive, send)

    def _find_error_handler(self, exception_type: Type[Exception]):
        for error_type, handler in self.route.__metadata__.error_handlers.items():
            if exception_type is error_type or issubclass(exception_type, error_type):
                return handler



import logging
from asyncio import Queue
from collections import defaultdict
from typing import Coroutine, Generic, Type, TypeVar

from wordlette.events import Observable
from wordlette.state_machines.predicates import always
from wordlette.state_machines.states import (
    State,
    InitialState,
    RequestCycle,
    Transition,
)
from wordlette.utils.maybe_awaitables import maybe_awaitable
from wordlette.utils.options import Option

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("StateMachine")

T = TypeVar("T")


class StoppedState(State[T]):
    async def enter_state(self):
        return


class StateMachine(Generic[T], Observable):
    def __init__(self, *states: Type[State] | Transition):
        self._states = self._build_state_graph(
            self._create_initial_state_transition(states[0]), *states
        )
        self._current_state = InitialState()
        self._stopped = True

        self._transition_stack = Queue()

    @property
    def state(self) -> State[T]:
        return self._current_state

    @property
    def started(self) -> bool:
        return not isinstance(self._current_state, InitialState)

    @property
    def stopped(self) -> bool:
        return self._stopped

    async def cycle(self):
        await self._queue_next_state()
        self._stopped = False
        while not self._transition_stack.empty():
            await self._exit_state()
            self._current_state = await self._transition_stack.get()
            logger.debug(f"Transitioning to {self._current_state}")
            await self._enter_state()

    async def _queue_next_state(self):
        match await self._get_next_state():
            case Option.Null():
                self._stopped = True
                await self._transition_stack.put(StoppedState())

            case Option.Value(constructor):
                await self._transition_stack.put(self._construct_state(constructor))

    async def _enter_state(self):
        match await maybe_awaitable(self._current_state.enter_state()):
            case RequestCycle():
                await self._queue_next_state()

    def _construct_state(self, constructor: Type[State[T]]) -> State[T]:
        state = constructor()
        if isinstance(state, Observable):
            state.__event_dispatch__.propagate_to(self.emit)

        return state

    def _exit_state(self) -> Coroutine[None, None, None]:
        return self._current_state.exit_state()

    def _build_state_graph(
        self, *states: Type[State] | Transition
    ) -> dict[Type[State[T]], dict]:
        graph = defaultdict(dict)
        for transition in states:
            match transition:
                case Transition(state, to_state, predicate):
                    graph[state][to_state] = predicate

                case type() as state if issubclass(state, State):
                    graph[state][StoppedState] = always

        return graph

    async def _get_next_state(self) -> Option[Type[State[T]]]:
        logger.debug(f"Finding next state from {self._current_state}")
        for next_state, predicate in self._states[type(self._current_state)].items():
            logger.debug(
                f"....Checking {next_state} when <{type(predicate).__name__}:{predicate.__name__}>"
            )
            if await predicate():
                return Option.Value(next_state)

        return Option.Null()

    def _create_initial_state_transition(
        self, state: Type[State[T]] | Transition
    ) -> Transition:
        match state:
            case Transition(state, _, _):
                return Transition(InitialState, state, always)

            case type() as state if issubclass(state, State):
                return Transition(InitialState, state, always)



from dataclasses import dataclass

from starlette.types import Scope

from wordlette.events import Event


class StartupEvent(Event):
    __event_name__ = "Startup"


@dataclass
class LifespanStartupEvent(Event):
    __event_name__ = "Lifespan.Startup"

    scope: Scope


@dataclass
class LifespanShutdownEvent(Event):
    __event_name__ = "Lifespan.Shutdown"

    scope: Scope



from bevy import inject, dependency

from wordlette.configs.managers import ConfigManager


class ConfigModel:
    __config_key__: str

    @classmethod
    @inject
    def __bevy_constructor__(cls, config: ConfigManager = dependency()):
        return config.get(cls.__config_key__, cls)



import logging
from pathlib import Path
from typing import Type, Sequence, TypeVar, Any, Callable

from wordlette.configs.handlers import ConfigHandler
from wordlette.core.exceptions import ConfigFileNotFound

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("Config")
T = TypeVar("T")


class ConfigManager:
    def __init__(self, handlers: Sequence[Type[ConfigHandler]] = ()):
        self._handlers: dict[str, ConfigHandler] = {}
        self._config: dict[str, Any] = {}

        self._register_handlers(handlers)

    @property
    def valid_extensions(self) -> set[str]:
        return set(self._handlers.keys())

    def get(
        self, key: str = "", constructor: Callable[[dict[str, Any]], T] | None = None
    ) -> T | Any:
        value = self._config[key] if key else self._config
        if constructor:
            value = constructor(**value)

        return value

    def find_config_file(self, name: str, directory: Path) -> Path | None:
        for extension in self._handlers:
            path = directory / f"{name}.{extension}"
            if path.exists():
                return path

    def load_config_file(self, name: str, directory: Path):
        path = self.find_config_file(name, directory)
        if not path:
            raise ConfigFileNotFound(
                f"Could not find any config files that matched the naming scheme '{name}.*' in '{directory}'."
            )

        handler = self._handlers[path.suffix[1:].casefold()]
        with path.open("rb") as open_file:
            self._config = handler.load(open_file)

        logger.debug(f"Loaded config file '{path}'.")

    def write_config_file(self, name: str, directory: Path, data: T):
        if path := self.find_config_file(name, directory):
            handler = self._handlers[path.suffix[1:].casefold()]
        else:
            extension = list(self._handlers)[0]
            handler = self._handlers[extension]
            path = directory / f"{name}.{extension}"

        self._config = data
        with path.open("wb") as open_file:
            handler.write(data, open_file)

    def add_handler(self, handler: Type[ConfigHandler]):
        for extension in handler.extensions:
            self._handlers[extension.casefold()] = handler()

    def remove_handler(self, handler: Type[ConfigHandler]):
        for extension in handler.extensions:
            del self._handlers[extension.casefold()]

    def _register_handlers(self, handlers: Sequence[Type[ConfigHandler]]):
        for handler in handlers:
            self.add_handler(handler)



from types import UnionType
from typing import Callable, get_args, Type, overload

from bevy import get_repository
from starlette.exceptions import HTTPException
from starlette.responses import Response, HTMLResponse
from starlette.routing import Router as _StarletteRouter
from starlette.types import Scope, Receive, Send

import wordlette.routes
from wordlette.requests import Request
from wordlette.utils.expand_dicts import from_dict


class StarletteRouter(_StarletteRouter):
    async def not_found(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "websocket":
            return await super().not_found(scope, receive, send)

        raise HTTPException(status_code=404)


class Router:
    def __init__(self):
        self.router = StarletteRouter()
        self.error_pages: dict[int, Callable[[int, Scope], Response]] = {}

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        try:
            return await self.router(scope, receive, send)
        except HTTPException as exc:
            page = self._get_error_page(exc.status_code, scope)
            await page(scope, receive, send)

    def add_error_page(
        self, status_code: int, get_page: Callable[[int, Scope], Response]
    ):
        self.error_pages[status_code] = get_page

    @overload
    def add_route(
        self,
        route: "type[wordlette.routes.Route]",
        include_in_schema: bool = True,
    ):
        ...

    @overload
    def add_route(
        self,
        path: str,
        *,
        route: Callable,
        methods: list[Type[Request]] | Type[Request] | UnionType = Request.Get,
        name: str = "",
        include_in_schema: bool = True,
    ):
        ...

    def add_route(self, path_or_route: "str | Type[wordlette.routes.Route]", **kwargs):
        match path_or_route:
            case str():
                self._add_new_route(
                    path_or_route,
                    *from_dict(kwargs).get_values(
                        route=...,
                        methods=Request.Get,
                        name="",
                        include_in_schema=True,
                    ),
                )

            case wordlette.routes.Route.matches_type() as route_type:
                self._add_route_type(
                    route_type, *from_dict(kwargs).get_values(include_in_schema=True)
                )

            case _:
                raise TypeError(f"{path_or_route!r} is not a valid route type")

    def _404_error_page(self, status_code: int, scope: Scope) -> Response:
        return HTMLResponse(
            f"<h1>{status_code}: Page Not Found</h1><p>Wordlette couldn't find any resources at <code>{scope['path']}</code>.<p>",
            status_code,
        )

    def _generic_error_page(self, status_code: int, scope: Scope) -> Response:
        return HTMLResponse(
            f"<h1>{status_code}: Error Encountered</h1><p>Wordlette encountered an error and couldn't recover.<p>",
            status_code,
        )

    def _add_route_type(
        self, route_type: "Type[wordlette.routes.Route]", include_in_schema: bool = True
    ):
        route = get_repository().get(route_type)
        self._add_new_route(
            route.path, route, route.methods.names, route.name, include_in_schema
        )

    def _add_new_route(
        self,
        path: str,
        route: Callable,
        methods: list[Type[Request]] | Type[Request] | UnionType,
        name: str,
        include_in_schema: bool,
    ):
        self.router.add_route(
            path,
            route,
            self._build_methods_list(methods),
            name or None,
            include_in_schema,
        )

    def _build_methods_list(
        self, methods: list[Type[Request] | str] | Type[Request] | str | UnionType
    ) -> list[str]:
        match methods:
            case [*values] if all(isinstance(value, str) for value in values):
                return [method.upper() for method in methods]

            case [*values] if all(issubclass(value, Request) for value in values):
                return [method.__name__.upper() for method in methods]

            case str() as method:
                return [method.upper()]

            case type() if issubclass(methods, Request):
                return [methods.__name__.upper()]

            case UnionType() as values:
                return self._build_methods_list(list(get_args(values)))

            case _:
                raise TypeError("Invalid type for methods")

    def _get_error_page(self, status_code: int, scope: Scope) -> Response:
        page = self._generic_error_page
        if status_code in self.error_pages:
            page = self.error_pages[status_code]

        elif 0 in self.error_pages:
            page = self.error_pages[0]

        elif status_code == 404:
            page = self._404_error_page

        return page(status_code, scope)



from pathlib import Path
from typing import Annotated

from bevy import get_repository

from wordlette.cms.extensions.error_pages import ErrorPages
from wordlette.cms.states.setup import Setup
from wordlette.core import WordletteApp
from wordlette.middlewares.router_middleware import RouterMiddleware
from wordlette.state_machines import StateMachine


def _setup_repository():
    repository = get_repository()
    repository.set(
        Annotated[Path, "package-resources"], Path(__file__).parent / "resources"
    )


_setup_repository()
app = WordletteApp(
    extensions=[ErrorPages],
    middleware=[RouterMiddleware],
    state_machine=StateMachine(Setup),
)



from io import StringIO
from typing import Any, BinaryIO

try:
    import yaml

except ImportError:
    yaml = None

from wordlette.configs.handlers import Serializable, ConfigHandler


class YamlHandler(ConfigHandler):
    extensions = {"yaml", "yml"}

    def load(self, file: BinaryIO) -> dict[str, Any]:
        if not yaml:
            raise ImportError("pyyaml is required to load YAML files.")

        return yaml.safe_load(file)

    def write(self, data: dict | Serializable, file: BinaryIO):
        if not yaml:
            raise ImportError("pyyaml is required to write YAML files.")

        def write(d):
            yaml.safe_dump(d, f := StringIO())
            file.write(f.getvalue().encode())

        match data:
            case dict():
                write(data)

            case Serializable():
                write(data.to_dict())

            case _:
                raise TypeError(
                    f"Cannot write type {type(data)!r} into a YAML file. Must be a dict or adhere to the "
                    f"`{Serializable.__module__}.{Serializable.__qualname__}` protocol."
                )

    @classmethod
    def supported(cls) -> bool:
        return yaml is not None



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar, NoReturn

from wordlette.core.exceptions import BaseWordletteException

E = TypeVar("E")
T = TypeVar("T")


class NoValueException(BaseWordletteException):
    pass


class Option(Generic[T], ABC):
    Value: "Type[Value[T]]"
    Null: "Type[Null[T]]"

    @property
    @abstractmethod
    def exception(self) -> BaseWordletteException | NoReturn:
        ...

    @abstractmethod
    def exception_or(self, default: E) -> BaseWordletteException | E:
        ...

    @property
    @abstractmethod
    def value(self) -> T | NoReturn:
        ...

    @abstractmethod
    def value_or(self, default: T) -> T:
        ...

    @abstractmethod
    def __bool__(self):
        ...


class Null(Option[T]):
    __match_args__ = ("exception",)

    def __init__(self, exception: BaseWordletteException | None = None):
        self._exception = exception

    @property
    def exception(self) -> BaseWordletteException:
        return self._exception or self._create_exception()

    def exception_or(self, _) -> BaseWordletteException:
        return self.exception

    @property
    def value(self) -> NoReturn:
        raise self._create_exception()

    def value_or(self, default: T) -> T:
        return default

    def __bool__(self):
        return False

    def _create_exception(self) -> NoValueException:
        exception = NoValueException("Null options have no value")
        if self._exception:
            exc_type = type(self._exception)
            exception.add_note(
                f"└─── caused by {exc_type.__module__}.{exc_type.__qualname__}: {self._exception}"
            )

        return exception


class Value(Option[T]):
    __match_args__ = ("value",)

    def __init__(self, value: T):
        self._value = value

    @property
    def exception(self) -> NoReturn:
        raise NoValueException("Value options have no exception")

    def exception_or(self, default: E) -> E:
        return default

    @property
    def value(self) -> T:
        return self._value

    def value_or(self, default: T) -> T:
        return self._value

    def __bool__(self):
        return True


Option.Value = Value
Option.Null = Null



from importlib import import_module
from pathlib import Path


class Extension:
    __match_args__ = ("name",)

    name: str

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "name"):
            cls.name = cls.__name__

    def __iter__(self):
        yield from (self.name, self)

    def __repr__(self):
        return f"<Extension:{self.name}>"


class PackageExtension(Extension):
    __slots__ = ("name", "package_import")

    def __init__(self, name: str, package_import: str):
        self.name = name
        self.package_import = package_import

    def load_extension(self):
        return import_module(self.package_import)


def scan_directory_for_public_modules(directory: Path) -> set[Path]:
    return {
        path
        for path in directory.iterdir()
        if (
            not path.name.startswith("_")
            and (path.suffix in {".py", ".pyc"} or path.is_dir())
        )
    }


def get_extensions_in_directory(directory: Path) -> list[Extension]:
    return [
        PackageExtension(path.stem, f"{path.parent.name}.{path.stem}")
        for path in scan_directory_for_public_modules(directory)
    ]



import json
from io import StringIO
from typing import Any, BinaryIO

from wordlette.configs.handlers import Serializable, ConfigHandler


class JsonHandler(ConfigHandler):
    extensions = {"json"}

    def load(self, file: BinaryIO) -> dict[str, Any]:
        return json.load(file)

    def write(self, data: dict | Serializable, file: BinaryIO):
        def write(d):
            json.dump(d, f := StringIO())
            file.write(f.getvalue().encode())

        match data:
            case dict():
                write(data)

            case Serializable():
                write(data.to_dict())

            case _:
                raise TypeError(
                    f"Cannot write type {type(data)!r} into a JSON file. Must be a dict or adhere to the "
                    f"`{Serializable.__module__}.{Serializable.__qualname__}` protocol."
                )



from pathlib import Path

from bevy import get_repository

from wordlette.configs.managers import ConfigManager
from wordlette.core.events import StartupEvent
from wordlette.events import Observer
from wordlette.extensions import Extension


def _create_config_manager():
    from wordlette.configs import JsonHandler, TomlHandler, YamlHandler

    manager = ConfigManager([JsonHandler])
    for handler in [TomlHandler, YamlHandler]:
        if handler.supported():
            manager.add_handler(handler)

    return manager


class Config(Extension, Observer):
    def __init__(self, config_manager: ConfigManager | None = None):
        super().__init__()
        self._config_manager = config_manager or _create_config_manager()
        self.config_file_stem = "wordlette.config"
        self.config_directory = Path.cwd()

        get_repository().set(ConfigManager, self._config_manager)

    async def on_startup(self, _: StartupEvent):
        self._config_manager.load_config_file(
            self.config_file_stem, self.config_directory
        )



def run():
    import sys

    match sys.argv:
        case [_, "serve"]:
            _start_server()

        case _:
            raise RuntimeError("Invalid command")


def _start_server():
    from wordlette.cms.app import app
    import uvicorn

    uvicorn.run(app, port=8000, log_level="info")



from typing import Type, Callable

from bevy import dependency
from starlette.responses import Response
from starlette.types import Scope, Send, Receive

from wordlette.events import Observer
from wordlette.middlewares import Middleware
from wordlette.routers import Router
from wordlette.routes import Route
from wordlette.utils.apply import apply


class RouteManager:
    def __init__(self):
        super().__init__()
        self._router: Router | None = None
        self._error_pages: dict[int, Callable[[int, Scope], Response]] = {}

    @property
    def router(self) -> Router:
        if not self._router:
            self.create_router()

        return self._router

    @router.setter
    def router(self, router):
        if not isinstance(router, Router):
            raise TypeError(f"router must be an instance of {Router.__qualname__}")

        self._router = router

    def add_error_page(
        self, status_code: int, get_page: Callable[[int, Scope], Response]
    ):
        self._error_pages[status_code] = get_page
        if self._router:
            self._router.add_error_page(status_code, get_page)

    def add_route(self, route: Type[Route]):
        self.router.add_route(route)

    def create_router(self, *routes: Type[Route]):
        self.router = Router()
        apply(self.add_route).to(routes)
        apply(self.add_error_page).to(self._error_pages.keys(), self._error_pages.values())


class RouterMiddleware(Middleware, Observer):
    route_manager: RouteManager = dependency()

    async def run(self, scope: Scope, receive: Receive, send: Send):
        await self.route_manager.router(scope, receive, send)



class BaseWordletteException(Exception):
    """Base exception for all wordlette exceptions."""

    pass


class MissingRoutePath(BaseWordletteException):
    """Raised when a route is missing a path."""


class NoRouteHandlersFound(BaseWordletteException):
    """Raised when a route has no registered method handlers."""


class InvalidExtensionOrConstructor(BaseWordletteException):
    """Raised when an invalid extension or constructor is given."""

    pass


class ConfigFileNotFound(BaseWordletteException):
    """Raised when a config file cannot be found."""

    pass



import asyncio
from typing import Type, Generator, TypeVar
from weakref import WeakSet

from wordlette.events.dispatch import Callback, Listener, EventDispatch
from wordlette.events.dispatchable import Dispatchable
from wordlette.events.events import Event
from wordlette.utils.contextual_methods import contextual_method

T = TypeVar("T")


class Observable(Dispatchable):
    """An object that can emit events."""

    __instances__: WeakSet["Observable"]
    __children__: WeakSet["Type[Observable]"]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.__event_dispatch__ = EventDispatch()
        cls.__children__ = WeakSet()
        cls.__instances__ = WeakSet()

        for parent in cls._get_dispatchable_bases():
            parent.__children__.add(cls)

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        cls.__instances__.add(instance)
        return instance

    @contextual_method
    def listen(obj, event: Type[Event], callback: Callback) -> Listener:
        """Adds a listener to the observable. When called as a class method this will register an event listener that
        listens for events emitted from any instance."""
        return obj.__event_dispatch__.listen(event, callback)

    @contextual_method
    def after(obj, event: Type[Event], callback: Callback) -> Listener:
        """Adds a listener to the observable that is called after the standard listeners run."""
        return obj.__event_dispatch__.after(event, callback)

    @contextual_method
    def before(obj, event: Type[Event], callback: Callback) -> Listener:
        """Adds a listener to the observable that is called before the standard listeners run."""
        return obj.__event_dispatch__.before(event, callback)

    @contextual_method
    async def emit(self, event: Event):
        """Emits an event to all listeners on the instances and propagates up through the superclasses."""
        await self.__event_dispatch__.emit(event)
        await type(self).__event_dispatch__.emit(event)
        await self._emit(self._get_dispatchable_supers(), event)

    @emit.classmethod
    async def emit(cls, event: Event):
        """Emits an event to all instances of the type, all child classes and their instances, and all super types."""
        await cls._emit_instances(event)
        await cls._emit_propagate_children(event)
        await cls._emit(cls._get_dispatchable_supers(), event)

    @contextual_method
    def stop(obj, event: Type[Event], callback: Callback):
        """Removes a listener from the observable."""
        obj.__event_dispatch__.stop(event, callback)

    @contextual_method
    def stop_after(obj, event: Type[Event], callback: Callback):
        """Removes a listener from the observable that is called after the standard listeners run."""
        obj.__event_dispatch__.stop_after(event, callback)

    @contextual_method
    def stop_before(obj, event: Type[Event], callback: Callback):
        """Removes a listener from the observable that is called before the standard listeners run."""
        obj.__event_dispatch__.stop_before(event, callback)

    @staticmethod
    def _emit(objs, event: Event):
        return asyncio.gather(*(obj.__event_dispatch__.emit(event) for obj in objs))

    @classmethod
    async def _emit_instances(cls, event: Event):
        await cls._emit(cls.__instances__, event)
        await asyncio.gather(
            *(child._emit_instances(event) for child in cls.__children__)
        )

    @classmethod
    async def _emit_propagate_children(cls, event):
        if cls.__children__:
            await asyncio.gather(
                *(child._emit_propagate_children(event) for child in cls.__children__)
            )

        await cls.__event_dispatch__.emit(event)

    @classmethod
    def _get_dispatchable_supers(
        cls, *, skip_self: bool = False
    ) -> Generator[Type[Dispatchable], None, None]:
        return cls._get_dispatchable_objects(cls.__mro__[1:], "__event_dispatch__")

    @classmethod
    def _get_dispatchable_bases(cls) -> "Generator[Type[Observable], None, None]":
        return cls._get_dispatchable_objects(cls.__bases__, "__children__")

    @staticmethod
    def _get_dispatchable_objects(objects, attr: str) -> Generator[Type[T], None, None]:
        return (obj for obj in objects if hasattr(obj, attr))



import logging

from bevy import inject, dependency
from starlette.types import Scope

from wordlette.cms.theming import Template, ThemeManager
from wordlette.core.events import LifespanStartupEvent
from wordlette.events import Observer
from wordlette.extensions import Extension
from wordlette.middlewares.router_middleware import RouteManager
from wordlette.utils.options import Value

logger = logging.getLogger(__name__)


class ErrorPages(Extension, Observer):
    def __init__(self):
        logger.debug("Initializing error pages")

    @inject
    async def on_startup(
        self, _: LifespanStartupEvent, route_manager: RouteManager = dependency()
    ):
        logger.debug("Adding error pages")
        route_manager.add_error_page(0, self._error_page)

    @inject
    def _error_page(self, status_code: int, scope: Scope, theme: ThemeManager = dependency()) -> Template:
        match theme.find_template(f"errors/{status_code}.html"):
            case Value():
                template_name = f"errors/{status_code}.html"

            case _:
                template_name = "errors/default.html"

        return Template(
            template_name,
            name="Page not found" if status_code == 404 else "An error was encountered",
            path=scope["path"],
            status_code=status_code,
        )



from typing import Type

from starlette.requests import Request as StarletteRequest


class Request(StarletteRequest):
    Delete: "Type[Delete]"
    Get: "Type[Get]"
    Head: "Type[Head]"
    Patch: "Type[Patch]"
    Post: "Type[Post]"
    Put: "Type[Put]"

    name: str

    def __init_subclass__(cls, **kwargs):
        cls.name = cls.__name__.upper()

    @classmethod
    def factory(cls, scope) -> "Request":
        match scope["method"].upper():
            case "DELETE":
                return cls.Delete(scope)

            case "GET":
                return cls.Get(scope)

            case "HEAD":
                return cls.Head(scope)

            case "PATCH":
                return cls.Patch(scope)

            case "POST":
                return cls.Post(scope)

            case "PUT":
                return cls.Put(scope)

            case _:
                return Request(scope)


class Delete(Request):
    ...


class Get(Request):
    ...


class Head(Request):
    ...


class Patch(Request):
    ...


class Post(Request):
    ...


class Put(Request):
    ...


Request.Delete = Delete
Request.Get = Get
Request.Head = Head
Request.Patch = Patch
Request.Post = Post
Request.Put = Put



import logging
from pathlib import Path
from typing import Annotated, Any, Mapping

import jinja2
from bevy import inject, dependency
from starlette.background import BackgroundTask
from starlette.responses import HTMLResponse, Response

from wordlette.cms.exceptions import ThemeNotFound, TemplateNotFound
from wordlette.utils.options import Option, Value, Null

logger = logging.getLogger("Theming")


class ThemeManager:
    wordlette_res: Annotated[Path, "package-resources"] = dependency()

    def __init__(self):
        self._theme = None
        self._secondary_themes = []
        self._jinja = jinja2.Environment(loader=JinjaTemplateLoader(self))

    @property
    def default_theme(self) -> Path:
        return self.wordlette_res / "themes" / "default"

    def find_template(self, template_name: str) -> Option[Path]:
        for theme in self._iterate_themes():
            template_location = theme / template_name
            if template_location.exists():
                return Value(template_location)

        return Null(
            TemplateNotFound(
                f"Could not find a template named {template_name!r} in any of the active themes."
            )
        )

    def set_theme(self, theme_location: Path):
        if not theme_location.exists():
            raise ThemeNotFound(
                f"Could not find a theme at {theme_location.resolve()}."
            )

        self._theme = theme_location
        logger.debug(f"Set theme to {theme_location.resolve()}")

    def add_secondary_theme(self, theme_location: Path):
        if not theme_location.exists():
            raise ThemeNotFound(
                f"Could not find a theme at {theme_location.resolve()} to use as a secondary theme."
            )

        self._secondary_themes.append(theme_location)

    def render_template(self, template_name: str, **context: Any) -> str:
        return self._jinja.get_template(template_name).render(**context)

    def _iterate_themes(self):
        if self._theme:
            yield self._theme

        yield from self._secondary_themes
        yield self.default_theme


class JinjaTemplateLoader(jinja2.BaseLoader):
    def __init__(self, theme_manager: ThemeManager):
        self.theme_manager = theme_manager

    def get_source(self, environment, template):
        match self.theme_manager.find_template(template):
            case Value(template_location):
                return self._load_template(template_location)

            case Null(exception):
                raise exception

    def _load_template(self, template_location: Path):
        logger.debug(
            f"Found template {template_location.name!r} at {template_location.resolve()}"
        )

        last_mtime = template_location.stat().st_mtime

        def is_current():
            return template_location.stat().st_mtime == last_mtime

        with template_location.open() as template_file:
            return template_file.read(), str(template_location), is_current


class Template(Response):
    def __init__(
        self,
        _template_name: str,
        _status_code: int = 200,
        *,
        _headers: Mapping[str, str] | None = None,
        _media_type: str | None = None,
        _background: BackgroundTask | None = None,
        **context: Any,
    ):
        self.name = _template_name
        self.context = context
        self.status_code = _status_code
        self._headers = _headers
        self.media_type = _media_type
        self.background = _background

    def __call__(self, *args, **kwargs):
        return HTMLResponse(
            self._render(),
            self.status_code,
            self.headers,
            self.media_type,
            self.background,
        )(*args, **kwargs)

    @inject
    def _render(self, theme_manager: ThemeManager = dependency()) -> str:
        return theme_manager.render_template(self.name, **self.context)



from types import EllipsisType
from typing import Any, ParamSpec, Iterable

P = ParamSpec("P")


class ExpandDict:
    """Takes any dict of str/value pairs and returns a callable that takes any number of keyword arguments. It returns a
    generator which yields all values of keys that match the keyword arguments. The value assigned to each keyword arg
    acts as a default value when the dict does not have a matching key. If the value is assigned Ellispsis/... then the
    key is required to be in the dict and a KeyError is raised when it is not found.

    Example:
        >>> a, b = ExpandDict({"a": 1}).get_values(a=..., b=2)
        >>> a, b
        (1, 2)
    """

    def __init__(self, dict_: dict[str, Any]):
        self.dict = dict_

    def get_values(self, **values) -> Iterable[Any]:
        return map(self._value_getter, values.items())

    def _value_getter(self, item: tuple[str, Any]) -> Any:
        match item:
            case (key, EllipsisType()):
                return self._get_required(self.dict, key)

            case (key, default):
                return self._get_optional(self.dict, key, default)

    @staticmethod
    def _get_optional(dict_: dict[str, Any], key: str, default: Any) -> Any:
        return dict_.get(key, default)

    @staticmethod
    def _get_required(dict_: dict[str, Any], key: str) -> Any:
        if key not in dict_:
            raise KeyError(
                f"Could not destructure the dict because it does not have a key {key!r}"
            )

        return dict_[key]


def from_dict(dict_: dict[str, Any]) -> ExpandDict:
    return ExpandDict(dict_)



from wordlette.core.exceptions import BaseWordletteException


class BaseWordletteCMSExecption(BaseWordletteException):
    """Base exception for all wordlette cms exceptions."""


class ThemeNotFound(BaseWordletteCMSExecption):
    """Raised when a theme cannot be found."""


class TemplateNotFound(BaseWordletteCMSExecption):
    """Raised when a template cannot be found."""



import asyncio
from collections import defaultdict
from functools import partial
from types import MethodType
from typing import Type, Callable, Awaitable, cast, TypeAlias, Any, Coroutine
from weakref import WeakMethod, ref

from wordlette.events.events import Event

Callback: TypeAlias = "Callable[[Event], Coroutine[Any, Any, None]] | Listener"
Registry: TypeAlias = dict[Type[Event], set[Callback]]
StopCallback: TypeAlias = Callable[[Callback], None]


class Listener:
    def __init__(self, callback: Callback, stop: StopCallback):
        self._callback = self._setup_callback(callback)
        self._stop = stop
        self._hash = hash(self.callback)

    def __call__(self, event: Event) -> Awaitable[None]:
        return self.callback(event)

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if callable(other):
            return self._hash == hash(other)

        raise TypeError(
            f"Cannot compare {self.__class__.__name__} to {other.__class__.__name__}"
        )

    @property
    def callback(self) -> Callable[[Event], Awaitable[None]]:
        return self._callback()

    def stop(self):
        self._stop(self)

    def _setup_callback(self, callback) -> ref[Callback]:
        weak_type = WeakMethod if isinstance(callback, MethodType) else ref
        return weak_type(callback, self._when_deleted)

    def _when_deleted(self, _: ref):
        self._stop(self)


class EventDispatch:
    """Dispatches events to listeners by type. All events can also be forwarded to an observer. There is support for
    applying listeners that run before or after the standard event listeners."""

    def __init__(self):
        self.after_listeners: Registry = defaultdict(set)
        self.before_listeners: Registry = defaultdict(set)
        self.listeners: Registry = defaultdict(set)
        self.observers: set[Callback] = set()

    def after(self, event: Type[Event], callback: Callback) -> Listener:
        """Registers an event listener that runs after the standard event listeners."""
        return self._register_listener(
            event, callback, self.after_listeners, self.stop_after
        )

    def before(self, event: Type[Event], callback: Callback) -> Listener:
        """Registers an event listener that runs before the standard event listeners."""
        return self._register_listener(
            event, callback, self.before_listeners, self.stop_before
        )

    async def emit(self, event: Event):
        """Emits an event to all listeners and observers."""
        event_type = type(event)
        listener_registries = (
            self.before_listeners[event_type],
            self.listeners[event_type],
            self.after_listeners[event_type],
            self.observers,
        )
        for listeners in listener_registries:
            await self._run_handlers(listeners, event)

    def listen(self, event: Type[Event], callback: Callback) -> Listener:
        """Registers an event listener."""
        return self._register_listener(event, callback, self.listeners, self.stop)

    def propagate_to(self, callback: Callback):
        """Adds an observer that will receive all events."""
        self.observers.add(callback)

    def stop(self, event: Type[Event], callback: Callback):
        """Stops an event listener."""
        self._stop_listener(event, callback, self.listeners)

    def stop_after(self, event: Type[Event], callback: Callback):
        """Stops an event listener that runs after the standard event listeners."""
        self._stop_listener(event, callback, self.after_listeners)

    def stop_before(self, event: Type[Event], callback: Callback):
        """Stops an event listener that runs before the standard event listeners."""
        self._stop_listener(event, callback, self.before_listeners)

    def _register_listener(
        self,
        event: Type[Event],
        callback: Callback,
        registry: Registry,
        stop: Callable[[Type[Event], Callback], None],
    ) -> Listener:
        listener = Listener(callback, partial(stop, event))
        registry[event].add(listener)
        return listener

    async def _run_handlers(self, listeners: set[Callback], event: Event):
        if not listeners:
            return

        return await asyncio.gather(*(listener(event) for listener in listeners))

    def _stop_listener(
        self, event: Type[Event], callback: Callback, registry: Registry
    ):
        if event not in registry:
            return

        if callback not in registry[event]:
            return

        registry[event].remove(cast(Listener, callback))



from wordlette.routes import Route


class SetupRoute(Route):
    """All subclasses of this abstract route type are automatically mounted by the Setup state."""

    class __metadata__:
        abstract = True
        registry = set()



from wordlette.events.dispatch import EventDispatch


class Dispatchable:
    __event_dispatch__: EventDispatch

    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls)
        instance.__event_dispatch__ = EventDispatch()
        return instance



from typing import TypeVar, Generic, Callable

T = TypeVar("T")


class Pipeline(Generic[T]):
    def __init__(self, *callables: Callable[[T], T]):
        self.callables = list(callables)

    def __call__(self, *args, **kwargs):
        return self.run(*args, **kwargs)

    def run(self, initial_value: T) -> T:
        value = initial_value
        for callable_ in self.callables:
            value = callable_(value)

        return value

    @staticmethod
    def no_params(func: Callable[[], T], *_) -> Callable[[T], T]:
        return lambda _: func()

    @staticmethod
    def with_params(
        func: Callable[[T], T], *args, ignore_value: bool = False, **kwargs
    ) -> Callable[[T], T]:
        def caller(value: T) -> T:
            _args = args if ignore_value else (value, *args)
            return func(*_args, **kwargs)

        return caller
